/*
 * Threaded hash object.
 * Note that this just has a second thread for hashing one block, then it blocks. 
 * We could chain the blocks together, but we don't.
 */

#ifndef THREADED_HASH_H
#define THREADED_HASH_H
#include <openssl/evp.h>
#ifdef HAVE_ERR_H
#include <err.h>
#endif

#include <sys/types.h>

/* threaded EVP hash object */
class threaded_hash {
private:;
#ifdef HAVE_PTHREAD
    pthread_mutex_t mutex;
#endif
    const EVP_MD *md;			// null if not valid
    EVP_MD_CTX ctx;			// hash context
    int hash_bytes;			// MD5 is 16
    u_char *buf;			// buffer being hashed in
    size_t bufsize;			// size of that buffer
    u_char *hashbuf;			// 
    char *hexbuf;			// 
    unsigned int be_threaded:1;
    static void *update_worker(void *arg){
	threaded_hash *t = (threaded_hash *)arg;
	EVP_DigestUpdate(&t->ctx,t->buf,t->bufsize);
	free(t->buf);
#ifdef HAVE_PTHREAD
	pthread_mutex_unlock(&t->mutex);
#endif
	return 0;
    }

public:;
    threaded_hash(const EVP_MD *md,unsigned int be_threaded){
	this->md          = md;
	if(md==0) return;		// invalid MD
#ifdef HAVE_PTHREAD
	if(pthread_mutex_init(&mutex,0)) err(1,"pthread_mutex");
#endif
	EVP_DigestInit(&ctx,md);
	this->hash_bytes  = EVP_MD_size(md);
	this->buf         = 0;
	this->bufsize     = 0;
	this->hashbuf     = (u_char *)calloc(hash_bytes,1);
	this->hexbuf      = (char *)calloc(hash_bytes*2+1,1);
	this->be_threaded = be_threaded>0 ? 1 : 0;
    }
    ~threaded_hash(){
#ifdef HAVE_PTHREAD
	pthread_mutex_destroy(&mutex);
#endif
	free(hashbuf);
	free(hexbuf);
    }

    void update(const u_char *buf,size_t bufsize){
	if(this->md==0) return;		// no MD set
	this->buf = (u_char *)malloc(bufsize);
	this->bufsize = bufsize;
	memcpy(this->buf,buf,bufsize);

#ifdef HAVE_PTHREAD
	pthread_t thread;
	pthread_mutex_lock(&mutex);
	if(be_threaded){
	    pthread_create(&thread,NULL,update_worker,(void *)this);
	    return;
	}
#endif
	update_worker((void *)this);
    }
    void final(u_char *mdbuf,unsigned int md_len){
	if(this->md==0) return;
#ifdef HAVE_PTHREAD
	pthread_mutex_lock(&mutex);
#endif
	EVP_DigestFinal(&ctx,mdbuf,&md_len);
#ifdef HAVE_PTHREAD
	pthread_mutex_unlock(&mutex);
#endif
    }
    u_char *hash(){
	if(this->md==0) return 0;
	final(hashbuf,hash_bytes);
	return hashbuf;
    }
    const char *hexhash(){
	if(this->md==0) return 0;
	hash();
	for(int i=0;i<hash_bytes;i++){
	    sprintf(hexbuf+i*2,"%02x",hashbuf[i]);
	}
	return hexbuf;
    }
    bool valid(){
	return this->md!=0;
    }

};
#endif
